<h3>Navigation</h3>
{{ toctree(maxdepth=3, collapse=True, includehidden=True) }}

<h3>Getting Started</h3>
<ul>
  <li><a href="{{ pathto('installation') }}">Installation</a></li>
  <li><a href="{{ pathto('quickstart') }}">Quickstart</a></li>
</ul>

<h3>API Reference</h3>
<ul>
  <li><a href="{{ pathto('py-modindex') }}">Module Index</a></li>
</ul>

<h3>Project Links</h3>
<ul>
  <li><a href="https://github.com/your-github-repo">GitHub Repository</a></li>
  <li><a href="https://pypi.org/project/quests/">PyPI Package</a></li>
</ul>
